#!/usr/env python

import argparse
import glob
import os
import os.path
import shutil
import sys
import tempfile

from update_package.buildcommon import *
from update_package.common import *
from livepatch_info import get_livepatch_info


def check_uuid(uuid):
    """
    Validate UUID
    """
    if not valid_uuid(uuid):
        sys.exit("%s: invalid UUID: %s" % (sys.argv[0], uuid))
    return uuid


def check_label(label):
    """
    Validate update label
    """
    if not valid_label(label):
        sys.exit("%s: invalid label: %s" % (sys.argv[0], label))
    return label


def check_guidance(guidance):
    """
    Validate after-apply guidance
    """
    if not valid_guidance(guidance):
        sys.exit("%s: invalid guidance: %s" % (sys.argv[0], guidance))
    return guidance


def check_uuid_label(uuid_label):
    """
    Validate uuid:label format
    """
    if not valid_uuid_label(uuid_label):
        sys.exit("%s: invalid uuid:label: %s" % (sys.argv[0], uuid_label))
    return uuid_label


def parse_args_or_exit(argv=None):
    """
    Parse command line options
    """
    parser = argparse.ArgumentParser(description="Build hotfix update")
    parser.add_argument("rpms", metavar="RPM", nargs="+",
                        help="Packages to include")
    parser.add_argument("--uuid", type=check_uuid, required=True,
                        help="Update unique identifier")
    parser.add_argument("-l", "--label", type=check_label, required=True,
                        help="Update textual label")
    parser.add_argument("-v", "--version", required=True,
                        help="Update version")
    parser.add_argument("-d", "--description", required=True,
                        help="Update description")
    parser.add_argument("-o", "--output", metavar="ISO", required=True,
                        help="Output ISO file")
    parser.add_argument("-k", "--key", metavar="UID",
                        help="UID of gpg key")
    parser.add_argument("--keyfile",
                        help="File name of gpg key")
    parser.add_argument("--base-requires",
                        help="Product version dependency")
    parser.add_argument("--after-apply", type=check_guidance,
                        help="Post apply guidance")
    parser.add_argument("--guidance-mandatory", action="store_true",
                        help="Guidance action to be immediate")
    parser.add_argument("--precheck", default=[], action="append",
                        help="Precheck script")
    parser.add_argument('--no-passphrase', action='store_false',
                        dest="prompt_pass",
                        help="Disable passphrase prompt")
    parser.add_argument("--requires", type=check_uuid_label,
                        metavar="UUID:Name-label", action="append",
                        help="Hotfix required to be already installed")
    parser.add_argument("--conflicts", type=check_uuid_label,
                        metavar="UUID:Name-label", action="append",
                        help="Hotfix conflict with current one")
    parser.add_argument("--rollsup", type=check_uuid_label,
                        metavar="UUID:Name-label", action="append",
                        help="Hotfix rollsup with current one")
    parser.add_argument("--properties", metavar="PROPERTY[,PROPERTY...]",
                        help="Hotfix properties")
    parser.add_argument("--stamps", metavar="TAR",
                        help="Archive containing rolled up stampfiles")
    parser.add_argument("--build-number", metavar="BRANCH/#",
                        help="Service Pack build number")
    parser.add_argument("--follows", type=check_label,
                        metavar="LABEL", action="append",
                        help="Service Pack that precedes this hotfix")
    parser.add_argument('--no-enforce-homogeneity', action='store_false',
                        dest="enforce_homogeneity",
                        help="Exclude this update from pool join checking")
    parser.add_argument('--update-stream',
                        help="Update belongs to a particular stream")

    return parser.parse_args(argv)


def create_livepatch_element_for_update_doc(work_dir):
    doc_name = os.path.join(work_dir, 'root/update.xml')
    livepatches = get_livepatch_info(work_dir)
    if len(livepatches):
        doc = xml.dom.minidom.parse(doc_name)
        top = doc.documentElement
        for livepatch in livepatches:
            desc = doc.createElement("livepatch")
            desc.setAttribute('component', livepatch['component'])
            if livepatch['complete-version']:
                desc.setAttribute('complete-version',
                                  livepatch['complete-version'])
            if livepatch['target-version']:
                desc.setAttribute('target-version',
                                  livepatch['target-version'])
            top.appendChild(desc)
        with open(doc_name, "w") as fileh:
            top.writexml(fileh, addindent="  ", newl='\n')


def main(argv):
    """
    Create and sign constituent parts then assemble ISO
    """
    args = parse_args_or_exit(argv)
    passphrase = None
    if args.key and args.prompt_pass:
        passphrase = get_gpg_key_pass(args.key)
    work_dir = tempfile.mkdtemp(prefix='build-hotfix-')
    try:
        os.mkdir(os.path.join(work_dir, 'root'))
        os.mkdir(os.path.join(work_dir, 'SOURCES'))
        control_pkg = build_control_rpm(work_dir, args.uuid, args.label,
                                        args.version, args.base_requires,
                                        args.precheck, args.requires,
                                        args.conflicts)
        local_rpms = [fetch_remote_file(rpm, work_dir) for rpm in args.rpms]
        size = calc_install_size(local_rpms)
        create_update_doc(os.path.join(work_dir, 'stampfile.xml'), args.uuid,
                          args.label, args.version, args.description,
                          args.after_apply, None, control_pkg, size,
                          args.guidance_mandatory, args.requires,
                          args.conflicts, args.rollsup, args.properties,
                          args.build_number,
                          enforce_homogeneity=args.enforce_homogeneity,
                          follows=args.follows, stream=args.update_stream)
        build_update_rpm(work_dir, args.uuid, args.label, args.version,
                         args.description, local_rpms, args.after_apply,
                         args.rollsup, args.stamps, args.build_number)
        create_update_doc(os.path.join(work_dir, 'root/update.xml'), args.uuid,
                          args.label, args.version, args.description,
                          args.after_apply, args.keyfile, control_pkg, size,
                          args.guidance_mandatory, args.requires,
                          args.conflicts, args.rollsup, args.properties,
                          args.build_number,
                          enforce_homogeneity=args.enforce_homogeneity,
                          follows=args.follows, stream=args.update_stream)
        copy_packages(work_dir, local_rpms)
        create_livepatch_element_for_update_doc(work_dir)
        if args.key:
            sign_update_doc(work_dir, args.key, passphrase)
            packages = glob.glob(os.path.join(work_dir, "root/Packages/*.rpm"))
            sign_packages(work_dir, args.key, passphrase, packages)
        groupfile = group_file(work_dir, args.label)
        create_repo(work_dir, args.key, passphrase, groupfile)
        build_iso(work_dir, args.output, args.uuid, args.label)
    finally:
        shutil.rmtree(work_dir, ignore_errors=True)


def _main():
    """
    Entry point for setuptools CLI wrapper
    """
    main(sys.argv[1:])


# Entry point when run directly
if __name__ == "__main__":
    _main()
