#############################################################################
#
# Generic Makefile rules for update generation.
#
# This file should be included at the bottom of an update makefile using
#
#     include $(shell update-resource --makefile)
#
# The following variables must be defined above that line:
#
#     JIRA_ID - the UPD project ticket, e.g. UPD-39
#
#     UPDATE_TYPE - one of 'servicePack', 'hotfix', 'supplementalPack' or
#                   'driverDisk'
#
#     UPDATE_UUID - the universal unique identifier, e.g.
#                   ae1d87ce-8896-432c-b58f-9aadc164e57e
#
#     UPDATE_LABEL - the textual reference, e.g. XS71E005
#
#     UPDATE_DESCRIPTION - the textual description, e.g.
#                          Public Availability: fixes to Storage
#
#     UPDATE_PACKAGES - the list of package names to be contained in the
#                       hotfix
#
# Any update makefile may define any of the following variables:
#
#     BASE_REQUIRES - either the platform-version or product-version constraint,
#                     e.g. platform-version = 2.1.0
#
#     UPDATE_GUIDANCE - the list of actions to perform following application,
#                       any of 'restartXAPI', 'restartHost', 'restartPV' or
#                       'restartHVM'
#
#     PRECHECK_SCRIPTS - the list of scripts to be executed during the update
#                       precheck operation
#
#     UPDATE_VERSION - the update version
#
#     UPDATE_STREAM - the update stream classification, 'premium'
#                     updates can only be applied to a host with
#                     Customer Success Services entitlement
#
#     EXTRA_BINARY_PACKAGES - the list of extra binary packages to be included
#                             in the additional binaries ISO
#
#     BINPKGS_VOLID - the volume label for the additional binaries ISO
#  
#     EXTRA_SOURCE_PACKAGES - the list of extra source packages to be included
#                             in the sources ISO
#
#     EXCLUDE_SOURCE_PACKAGES - the list of source packages to be excluded from
#                               the sources ISO
#
#     EXTRA_BUILD_OPTIONS - additional arguments to build-update
#
# A hotfix makefile may define any of the following variables:
#
#     HOTFIX_ROLLSUP - the list of hotfix labels this hotfix includes the
#                      content of
#
#     HOTFIX_GUIDANCE_MANDATORY - 'true' if guidance actions must be performed
#                                 immediately after application
#
#     HOTFIX_REQUIRES - the list of hotfix labels that must be present before
#                       this hotfix is applied
#
#     HOTFIX_CONFLICTS - the list of hotfix labels that must not be present
#                        when this hotfix is applied
#
#     HOTFIX_PROPERTIES - the list property labels
#
# A supplementalPack makefile may define any of the following variables:
#
#     REMOVE_SCRIPT - the script to be executed during the update remove
#                     operation
#
# A servicePack makefile must define the following variables:
#
#     BUILD_NUMBER - the branch and build number
#
#############################################################################

OUTPUT_DIR ?= $(abspath ../output)
WORK_DIR := /tmp/$(UPDATE_LABEL)-build

# Generated outputs
ifeq ($(UPDATE_TYPE),driverDisk)
ISO_PREFIX := $(UPDATE_LABEL)-$(DRIVER_VERSION)
else
ISO_PREFIX := $(UPDATE_LABEL)
endif

ISO := $(OUTPUT_DIR)/$(ISO_PREFIX).iso
BINPKGS_ISO := $(OUTPUT_DIR)/$(ISO_PREFIX)-binpkgs.iso
SRC_ISO := $(OUTPUT_DIR)/$(ISO_PREFIX)-sources.iso
FULL_SRC_ISO := $(OUTPUT_DIR)/$(ISO_PREFIX)-full-sources.iso
PUBKEY = $(OUTPUT_DIR)/$(notdir $(SIGN_KEYFILE))

UPDATE_REPOS ?= xenserver-specs
UPDATE_BUILD_REPOS ?= xenserver-specs
ifeq ($(UPDATE_TYPE),$(filter $(UPDATE_TYPE),hotfix supplementalPack))
BASE_REQUIRES ?= product-version = $$product_version
endif
ifeq ($(UPDATE_TYPE),driverDisk)
BASE_REQUIRES ?= kernel-uname-r = $$kernel_uname_r
endif
UPDATE_VERSION ?= 1.0
PRECHECK_SCRIPTS ?= $(PRECHECK_SCRIPT)
BINPKGS_VOLID ?= $(UPDATE_LABEL) Packages
SOURCE_VOLID ?= $(UPDATE_LABEL) Sources
FULL_SOURCE_VOLID ?= $(UPDATE_LABEL) Full Sources

SIGN_KEYFILE ?= ~/RPM-GPG-KEY-XS-Eng-Local
SIGN_UID = $(shell gpg --with-fingerprint $(SIGN_KEYFILE) | sed -ne 's/^pub  [^ ]* [^ ]* \(.*\)/\1/p')

UPDATE_UUID_LABEL:=$(UPDATE_UUID):$(UPDATE_LABEL)

BUILD_OPTIONS = --uuid $(UPDATE_UUID) --label $(UPDATE_LABEL) --version $(UPDATE_VERSION)
BUILD_OPTIONS += --description "$(UPDATE_DESCRIPTION)"
BUILD_OPTIONS += $(UPDATE_GUIDANCE:%=--after-apply %)
BUILD_OPTIONS += --key "$(SIGN_UID)" --keyfile $(SIGN_KEYFILE) --no-passphrase
BUILD_OPTIONS += $(PRECHECK_SCRIPTS:%=--precheck %)
BUILD_OPTIONS += $(UPDATE_STREAM:%=--update-stream %)
ifneq ($(BASE_REQUIRES),)
BUILD_OPTIONS += --base-requires "$(BASE_REQUIRES)"
endif
BUILD_OPTIONS += $(EXTRA_BUILD_OPTIONS)

BUILD_CMD = build-update

ifeq ($(UPDATE_TYPE),supplementalPack)
BUILD_OPTIONS += $(REMOVE_SCRIPT:%=--remove %)
else
ifeq ($(UPDATE_TYPE),$(filter $(UPDATE_TYPE),hotfix servicePack))
BUILD_OPTIONS += $(HOTFIX_GUIDANCE_MANDATORY:%=--guidance-mandatory)
BUILD_OPTIONS += $(HOTFIX_PROPERTIES:%=--properties %)
BUILD_OPTIONS += $(call hotfix_params,--rollsup,$(COMPLETE_ROLLUPS))
BUILD_OPTIONS += $(call hotfix_params,--requires,$(HOTFIX_REQUIRES))
BUILD_OPTIONS += $(call hotfix_params,--conflicts,$(HOTFIX_CONFLICTS))
BUILD_CMD = build-hotfix-update
endif
ifeq ($(UPDATE_TYPE),hotfix)
BUILD_OPTIONS += $(HOTFIX_ROLLSUP:%=--stamps $(STAMPFILES_TAR))
BUILD_OPTIONS += $(HOTFIX_FOLLOWS:%=--follows %)
endif
ifeq ($(UPDATE_TYPE),servicePack)
BUILD_OPTIONS += $(BUILD_NUMBER:%=--build-number %)
endif
endif

ifdef NO_SRC_REPOS
FETCH_PKG_SRC_OPTS := --no-sources "$(NO_SRC_REPOS)"
endif

PACKAGES_LIST := $(WORK_DIR)/$(UPDATE_LABEL)-packages-list
PACKAGE_PATHS := $(WORK_DIR)/package-paths
PACKAGES_DIR := $(WORK_DIR)/Packages
BINARY_PACKAGES_LIST := $(WORK_DIR)/binary-packages-list
BINARY_PACKAGE_PATHS := $(WORK_DIR)/binary-package-paths
BINARY_PACKAGES_DIR := $(WORK_DIR)/BinaryPackages
SOURCE_PACKAGES_PATHS := $(WORK_DIR)/source-packages-paths
FULL_SOURCE_PACKAGES_PATHS := $(WORK_DIR)/full-source-packages-paths
BINPKGS_SOURCE_PACKAGES_PATHS := $(WORK_DIR)/direct-source-packages-paths
SOURCE_PACKAGES_DIR := $(WORK_DIR)/SourcePackages
RELEASE_VERSIONS := $(WORK_DIR)/release-versions
UUID_FILE := $(WORK_DIR)/$(UPDATE_LABEL)-uuid
BINPKGS_ISO_DIR := $(WORK_DIR)/binpkgs-iso
SOURCE_ISO_DIR := $(WORK_DIR)/source-iso
FULL_SOURCE_ISO_DIR := $(WORK_DIR)/full-source-iso
UPDATE_DOC := $(WORK_DIR)/$(UPDATE_LABEL)-update.xml
COMPLETE_ROLLUPS_LIST := $(WORK_DIR)/$(UPDATE_LABEL)-complete-rollups-list
STAMPFILES_TAR := $(WORK_DIR)/stampfiles.tar
GROUP_FILE := $(WORK_DIR)/groups.xml
SRC2PKGMAP := $(WORK_DIR)/source-to-binary-map

PAYLOAD_PACKAGES = $$(<$(PACKAGES_LIST))
COMPLETE_ROLLUPS = $(shell cat $(COMPLETE_ROLLUPS_LIST))
ALL_RELATED = $(COMPLETE_ROLLUPS) $(HOTFIX_REQUIRES) $(HOTFIX_CONFLICTS)
ISO_TARGET_FILES = $(PACKAGE_PATHS) $(RELEASE_VERSIONS) $(SIGN_KEYFILE) $(ALL_RELATED:%=$(WORK_DIR)/%-uuid) $(STAMPFILES_TAR) $(PRECHECK_SCRIPT)

ifeq ($(UPDATE_TYPE),driverDisk)
BUILD_OPTIONS += --groupfile $(GROUP_FILE)
ISO_TARGET_FILES += $(GROUP_FILE)
endif

# macros to expand a name-label to "UUID:name-label"
hotfix_param = $(shell echo $(1) $$(<$(WORK_DIR)/$(2)-uuid):$(2))
hotfix_params = $(foreach h,$(2),$(call hotfix_param,$(1),$(h)))

# macro to create an empty directory
mkdir_empty = rm -rf $(1); mkdir -p $(1)

# macro to prepend an arg to each element of a list
arg_builder = $(foreach p,$(1),$(2) $(p) )

GENISOIMAGE := genisoimage -joliet -joliet-long -rational-rock
LIST_BINPKGS := list-binary-packages
REPOQUERY = repoquery '--disablerepo=*' $(UPDATE_BUILD_REPOS:%=--enablerepo=%)
YUMDOWNLOADER = yumdownloader '--disablerepo=*' $(UPDATE_REPOS:%=--enablerepo=%)
YUMDOWNLOADER_OPTIONS ?= --exclude '*.i?86'

ERROR_FILTER := tee /dev/stderr | (! grep -E "^No Match for argument |^Dependency resolution failed|No more mirrors to try")
error_filter = grep -q -E "^Error getting repository data for |^No Match for argument |^Dependency resolution failed|No more mirrors to try" $(1) && exit 1 || :

.PHONY: build
build: $(ISO) $(BINPKGS_ISO) $(SRC_ISO) $(PUBKEY)

.PHONY: clean
clean:
	rm -rf $(WORK_DIR)

.PHONY: has-xen-livepatch
has-xen-livepatch:
ifeq ($(UPDATE_TYPE),hotfix)
	@echo "$(LIVEPATCH_PACKAGES)" | grep --quiet 'xen-\([[:alnum:]]*-\)\?livepatch_' && echo true || echo false
else
	echo false
endif

.PHONY: has-kernel-livepatch
has-kernel-livepatch:
ifeq ($(UPDATE_TYPE),hotfix)
	@echo "$(LIVEPATCH_PACKAGES)" | grep --quiet 'kernel-\([[:alnum:]]*-\)\?livepatch_' && echo true || echo false
else
	echo false
endif

$(ISO): $(ISO_TARGET_FILES)
	@mkdir -p $(@D)
	source $(RELEASE_VERSIONS); \
	$(BUILD_CMD) --output $@ $(BUILD_OPTIONS) $$(< $(PACKAGE_PATHS))

$(BINPKGS_ISO): $(BINARY_PACKAGE_PATHS)
	$(call mkdir_empty,$(BINPKGS_ISO_DIR)/Packages)
	[ ! -s $(BINARY_PACKAGE_PATHS) ] || cp -p $$(<$(BINARY_PACKAGE_PATHS)) $(BINPKGS_ISO_DIR)/Packages
	cd $(BINPKGS_ISO_DIR); createrepo .
	$(GENISOIMAGE) -V "$(BINPKGS_VOLID)" -o $@ $(BINPKGS_ISO_DIR)

$(SRC_ISO): $(SOURCE_PACKAGES_PATHS)
	@mkdir -p $(@D)
	$(call mkdir_empty,$(SOURCE_ISO_DIR)/SPackages)
	[ ! -s $(SOURCE_PACKAGES_PATHS) ] || cp -p $$(<$(SOURCE_PACKAGES_PATHS)) $(SOURCE_ISO_DIR)/SPackages
	cd $(SOURCE_ISO_DIR); createrepo .
	$(GENISOIMAGE) -V "$(SOURCE_VOLID)" -o $@ $(SOURCE_ISO_DIR)

$(FULL_SRC_ISO): $(FULL_SOURCE_PACKAGES_PATHS)
	@mkdir -p $(@D)
	$(call mkdir_empty,$(FULL_SOURCE_ISO_DIR)/SPackages)
	[ ! -s $(FULL_SOURCE_PACKAGES_PATHS) ] || cp -p $$(<$(FULL_SOURCE_PACKAGES_PATHS)) $(FULL_SOURCE_ISO_DIR)/SPackages
	cd $(FULL_SOURCE_ISO_DIR); createrepo .
	$(GENISOIMAGE) -V "$(FULL_SOURCE_VOLID)" -o $@ $(FULL_SOURCE_ISO_DIR)

$(PUBKEY): $(SIGN_KEYFILE)
	cp -p $^ $@

$(PACKAGES_LIST): $(HOTFIX_ROLLSUP:%=$(WORK_DIR)/%-packages-list)
	@mkdir -p $(@D)
	cat $^ /dev/null >$@.unsorted
	shopt -s nullglob; for f in $(UPDATE_PACKAGES); do echo "$$f" >>$@.unsorted; done
	sort -u -o $@ $@.unsorted

.DELETE_ON_ERROR: $(PACKAGE_PATHS)
$(PACKAGE_PATHS): $(PACKAGES_LIST) $(EXTRA_PACKAGE_PATHS)
	@mkdir -p $(PACKAGES_DIR)
	[ ! -s $(PACKAGES_LIST) ] || ( \
		echo "Downloading packages for:"; \
		cat $(PACKAGES_LIST); \
		trap "rm -f /tmp/yum-out.$$$$" EXIT; \
		$(YUMDOWNLOADER) $(YUMDOWNLOADER_OPTIONS) --destdir $(PACKAGES_DIR) $(PAYLOAD_PACKAGES) 2>&1 | tee /tmp/yum-out.$$$$; \
		$(call error_filter,/tmp/yum-out.$$$$); \
	)
	( shopt -s nullglob; for f in $(PACKAGES_DIR)/*.rpm; do echo "$$f"; done ) >$@
	shopt -s nullglob; for f in $(EXTRA_PACKAGE_PATHS); do echo "$$f" >>$@; done

$(BINARY_PACKAGE_PATHS): $(BINARY_PACKAGES_LIST)
	@mkdir -p $(BINARY_PACKAGES_DIR)
	[ ! -s $(BINARY_PACKAGES_LIST) ] || ( \
		echo "Downloading packages for:"; \
		cat $(BINARY_PACKAGES_LIST); \
		trap "rm -f /tmp/yum-out.$$$$" EXIT; \
		$(YUMDOWNLOADER) $(YUMDOWNLOADER_OPTIONS) --destdir $(BINARY_PACKAGES_DIR) $$(<$(BINARY_PACKAGES_LIST)) 2>&1 | tee /tmp/yum-out.$$$$; \
		$(call error_filter,/tmp/yum-out.$$$$); \
	)
	shopt -s nullglob; for f in $(BINARY_PACKAGES_DIR)/*.rpm; do echo "$$f" >>$@; done

$(BINARY_PACKAGES_LIST): $(BINPKGS_SOURCE_PACKAGES_PATHS)
	$(REPOQUERY) --qf '%{sourcerpm} %{name}' '*-devel' '*-debuginfo' >$(SRC2PKGMAP)
	( [ ! -s $(BINPKGS_SOURCE_PACKAGES_PATHS) ] || $(LIST_BINPKGS) $(BINPKGS_SOURCE_PACKAGES_PATHS) $(SRC2PKGMAP) ) >$@
	shopt -s nullglob; for f in $(EXTRA_BINARY_PACKAGES); do echo "$$f" >>$@; done

$(BINPKGS_SOURCE_PACKAGES_PATHS): $(PACKAGES_LIST)
	$(REPOQUERY) --nvr --srpm $(PAYLOAD_PACKAGES) | sort -u | awk '{print "$(SOURCE_PACKAGES_DIR)/"$$1".src.rpm"}' >$@

.DELETE_ON_ERROR: $(SOURCE_PACKAGES_PATHS)
$(SOURCE_PACKAGES_PATHS): $(FULL_SOURCE_PACKAGES_PATHS)
	for f in $$(<$(FULL_SOURCE_PACKAGES_PATHS)); do echo -n "$$f^" ; rpm -q --nosignature --qf '%{license}\n' -p $$f; done | awk -v IGNORECASE=1 -F^ '$$2!~"proprietary" && $$2!~"nvidia" {print $$1}' >$@

$(FULL_SOURCE_PACKAGES_PATHS): $(PACKAGES_LIST)
	$(mkdir_empty) $(SOURCE_PACKAGES_DIR)
	[ ! -s "$(PACKAGES_LIST)" ] || fetch-package-sources --output $(SOURCE_PACKAGES_DIR) $(FETCH_PKG_SRC_OPTS) '$(UPDATE_BUILD_REPOS)' '$(UPDATE_REPOS)' $(PAYLOAD_PACKAGES)
	shopt -s nullglob; for f in $(SOURCE_PACKAGES_DIR)/*.rpm; do echo "$$f" >>$@; done
	shopt -s nullglob; for f in $(EXTRA_SOURCE_PACKAGE_PATHS); do echo "$$f" >>$@; done

$(WORK_DIR)/rollup.mk: $(COMPLETE_ROLLUPS_LIST)
	mkdir -p $(@D)
	awk '{ print "$(ISO): $(WORK_DIR)/"$$1"-uuid" }' $< >$@
	awk '{ print "$(STAMPFILES_TAR): $(WORK_DIR)/"$$1"-update.xml" }' $< >>$@

$(COMPLETE_ROLLUPS_LIST): $(HOTFIX_ROLLSUP:%=$(WORK_DIR)/%-complete-rollups-list)
	mkdir -p $(@D)
	cat $^ /dev/null >$@.unsorted
	shopt -s nullglob; for f in $(HOTFIX_ROLLSUP); do echo "$$f" >>$@.unsorted; done
	sort -u -o $@ $@.unsorted

$(STAMPFILES_TAR):
	mkdir -p $(@D)
	touch $(WORK_DIR)/empty
	tar cvf $@ --transform=s+.*/++ $(COMPLETE_ROLLUPS:%=$(WORK_DIR)/%-update.xml) $(WORK_DIR)/empty

.DELETE_ON_ERROR: $(RELEASE_VERSIONS)
$(RELEASE_VERSIONS):
	@mkdir -p $(@D)
	-( yum -q repolist all xenserver-specs | grep -q '^.\?xenserver-specs' && \
		repoquery --provides kernel xenserver-release | awk '$$1!~/[()]/ { gsub("-","_",$$1); print $$1"=\""$$3"\"" }' ) >$@
	@echo "Release versions:"
	@cat $@

$(SIGN_KEYFILE):
	@echo "Generating local signing key..."
	[ -d ~/.gnupg ] || gpg -k
	manage-signing-keys --create "Local signing key" --expires 3m --no-comment --no-email --export $@

# Rule to create a file containing the UUID of this hotfix
.DELETE_ON_ERROR: $(UUID_FILE)
$(UUID_FILE):
	echo $(UPDATE_UUID) >$@

# Rule to create a minimal update XML document for this hotfix
$(UPDATE_DOC):
	create-update-doc --uuid $(UPDATE_UUID) --label $(UPDATE_LABEL) --version $(UPDATE_VERSION) --description "$(UPDATE_DESCRIPTION)" --rolled-up-by $(ROLLED_UP_BY) $(HOTFIX_FOLLOWS:%=--follows %) --output $@

# Rule to create group file for driver disk which contains drivers group
$(GROUP_FILE):
	$(info "Creating group file for driver disks update...")
	create-group-file  --label $(UPDATE_LABEL) --driver_packages $(DRIVER_PACKAGES) --output $@

# Rule to create a file containing the UUID of a hotfix related to this one
$(WORK_DIR)/%-uuid:
	$(MAKE) -C ../$* $@ WORK_DIR=$(WORK_DIR)

# Rule to create a file containing the package names in a hotfix rolled up by this one
$(WORK_DIR)/%-packages-list:
	$(MAKE) -C ../$* $@ WORK_DIR=$(WORK_DIR) UPDATE_TYPE=$(UPDATE_TYPE)

# Rule to create the full list of hotfixes rolled up by this one
$(WORK_DIR)/%-complete-rollups-list:
	$(MAKE) -C ../$* $@ WORK_DIR=$(WORK_DIR) UPDATE_TYPE=$(UPDATE_TYPE)

# Rule to create the minimal update XML document for a hotfix rolled up by this one
$(WORK_DIR)/%-update.xml:
	$(MAKE) -C ../$* $@ WORK_DIR=$(WORK_DIR) ROLLED_UP_BY=$(UPDATE_UUID_LABEL)

-include $(WORK_DIR)/rollup.mk
